"""
Pharmalyzer - ADME and Toxicity Screening Toolkit

Author: [Sorour Hassani]
Email: s.hassani@alum.semnan.ac.ir & sorour.hasani@gmail.com
License: MIT
"""

def min_max_scale(data, column):
    values = [row[column] for row in data if isinstance(row[column], (int, float))]
    if not values:
        return data
    min_val, max_val = min(values), max(values)
    range_val = max_val - min_val
    if range_val == 0:

       for row in data:
           if isinstance(row[column], (int, float)):
              row[column] = (row[column] - min_val) / range_val
    return data

def min_max_scale_multiple(data, columns):
    for col in columns:
        data = min_max_scale(data, col)
    return data



"""
Pharmalyzer - ADME and Toxicity Screening Toolkit

Author: [Sorour Hassani]
Email: s.hassani@alum.semnan.ac.ir & sorour.hasani@gmail.com
License: MIT
"""
# Pharmalyzer/qsar.py

import os
import pickle
import math
from rdkit import Chem
from rdkit.Chem import Descriptors, rdMolDescriptors
from sklearn.ensemble import RandomForestRegressor
from sklearn.model_selection import train_test_split
from sklearn.metrics import mean_squared_error, r2_score

class QSARModel:
    """
    A generic QSAR trainer/predictor for multiple numeric properties.

    Given:
      - a list of dicts (or DataFrame rows) where each row has:
          'SMILES', <prop_1>, <prop_2>, ..., <prop_n>
      - a directory path to save models as '{property_name}_model.pkl'
    It will:
      - compute a fixed set of descriptors for each SMILES
      - for each property column, train a separate RandomForestRegressor
      - save each trained model as a .pkl file under the specified models directory
      - expose a `predict` method to load a saved model and predict for a new SMILES
    """

    def __init__(self, models_dir="qsar_models"):
        """
        Args:
            models_dir (str): folder where trained model pickles will be written/read.
        """
        self.models_dir = models_dir
        os.makedirs(self.models_dir, exist_ok=True)

    def compute_descriptors(self, smiles):
        """
        Compute a fixed set of descriptors from a SMILES string.
        Returns a list of descriptor values in consistent order.
        """
        mol = Chem.MolFromSmiles(smiles)
        if mol is None:
            return None

        mw = Descriptors.MolWt(mol)
        logp = Descriptors.MolLogP(mol)
        tpsa = rdMolDescriptors.CalcTPSA(mol)
        hbd = Descriptors.NumHDonors(mol)
        hba = Descriptors.NumHAcceptors(mol)
        rot_bonds = Descriptors.NumRotatableBonds(mol)
        ring_count = mol.GetRingInfo().NumRings()
        heavy_atoms = mol.GetNumHeavyAtoms()

        return [
            mw,
            logp,
            tpsa,
            hbd,
            hba,
            rot_bonds,
            ring_count,
            heavy_atoms
        ]

    def prepare_dataset(self, rows, property_name):
        """
        Given an iterable of dict-like rows, extract:
          - X: list of descriptor lists
          - y: list of target values for property_name
        Skip any row where descriptors fail or property is missing/non-numeric.

        Args:
            rows: list of dicts (or DataFrame rows) with keys 'SMILES' and property_name
            property_name (str): column/key indicating the numeric target variable

        Returns:
            X_list (list of descriptor lists), y_list (list of floats)
        """
        X_list = []
        y_list = []
        for r in rows:
            smi = r.get("SMILES")
            if smi is None:
                continue
            desc = self._compute_descriptors(smi)
            if desc is None:
                continue
            if any(math.isnan(d) for d in desc):
                 print(f"[WARN] Skipping SMILES due to NaN descriptors: {smi}")
                 continue
            val = r.get(property_name)
            try:
                val_f = float(val)
            except (ValueError, TypeError):
                continue
            X_list.append(desc)
            y_list.append(val_f)
        return X_list, y_list

    def train_and_save(self, dataset_rows, property_names, test_size=0.2, random_state=42):
        """
        For each property in property_names, train a RandomForestRegressor
        and save the model to self.models_dir + "/{prop}_model.pkl".

        Args:
            dataset_rows (list of dicts): each dict must have 'SMILES' plus all properties
            property_names (list of str): list of column names to train on
            test_size (float): fraction for test split
            random_state (int): for reproducibility

        Returns:
            report (dict): for each property, a dict with train/test metrics
        """
        report = {}
        for prop in property_names:
            X, y = self._prepare_dataset(dataset_rows, prop)
            if len(X) < 10:
                report[prop] = {
                    "status": "skipped (insufficient data)",
                    "n_samples": len(X)
                }
                continue

            X_train, X_test, y_train, y_test = train_test_split(
                X, y, test_size=test_size, random_state=random_state
            )

            model = RandomForestRegressor(n_estimators=100, random_state=random_state)
            model.fit(X_train, y_train)

            preds = model.predict(X_test)
            mse = mean_squared_error(y_test, preds)
            r2 = r2_score(y_test, preds)

            model_filename = os.path.join(self.models_dir, f"{prop}_model.pkl")
            with open(model_filename, "wb") as fout:
                pickle.dump(model, fout)

            report[prop] = {
                "status": "trained",
                "n_samples": len(X),
                "mse_test": mse,
                "r2_test": r2,
                "model_file": model_filename
            }

        return report

    def predict(self, property_name, smiles_list):
        """
        Load the saved model for property_name and predict values for each SMILES in smiles_list.

        Args:
            property_name (str): which property’s model to load (must match training name)
            smiles_list (list of str): list of SMILES strings to predict

        Returns:
            A list of predicted floats (same length as smiles_list). If a SMILES fails to parse,
            the corresponding output will be None.
        """
        model_path = os.path.join(self.models_dir, f"{property_name}_model.pkl")
        if not os.path.isfile(model_path):
            raise FileNotFoundError(f"Model file not found: {model_path}")

        with open(model_path, "rb") as fin:
            model = pickle.load(fin)

        preds = []
        for smi in smiles_list:
            desc = self._compute_descriptors(smi)
            if desc is None:
                preds.append(None)
            else:
                preds.append(float(model.predict([desc])[0]))
        return preds




"""
Pharmalyzer - ADME and Toxicity Screening Toolkit

Author: [Sorour Hassani]
Email: s.hassani@alum.semnan.ac.ir & sorour.hasani@gmail.com
License: MIT
"""

import csv
import math

def rank_features_by_variance(matrix, headers, descending=True, show_top=True, top_k=None, save_path=None, return_indices=False):
    """
    Compute and rank features by variance with extended options.

    Parameters:
        matrix (List[List[float]]): 2D dataset (rows of samples).
        headers (List[str]): Column names.
        descending (bool): Sort from high to low if True.
        show_top (bool): Print most/least variant features if True.
        top_k (int or None): Number of top features to return. If None, return all.
        save_path (str or None): If specified, save results to this CSV file.
        return_indices (bool): If True, return feature indices instead of names.

    Returns:
        List[Tuple[str or int, float]]: Sorted list of (feature, variance).
    """

    def compute_variance(values):
        n = len(values)
        if n == 0:
            return 0.0
        mean = sum(values) / n
        return sum((x - mean) ** 2 for x in values) / n

    if not matrix or not headers:
        return []

    num_cols = len(headers)
    columns = [[] for _ in range(num_cols)]

    for row in matrix:
        for i in range(min(len(row), num_cols)):
            try:
                val = float(row[i])
                if not math.isnan(val):
                   columns[i].append(float(row[i]))
            except (ValueError, TypeError):
                continue

    variances = {
        i if return_indices else headers[i]: compute_variance(columns[i])
        for i in range(num_cols)
        if columns[i]
    }

    sorted_features = sorted(variances.items(), key=lambda x: x[1], reverse=descending)

    if show_top and sorted_features:
        print(f"🔺 Highest variance: {sorted_features[0][0]} = {sorted_features[0][1]:.4f}")
        print(f"🔻 Lowest variance: {sorted_features[-1][0]} = {sorted_features[-1][1]:.4f}")

    if top_k is not None:
        sorted_features = sorted_features[:top_k]

    if save_path:
        with open(save_path, 'w', newline='') as csvfile:
            writer = csv.writer(csvfile)
            writer.writerow(['Feature', 'Variance'])
            writer.writerows(sorted_features)

    return sorted_features



"""
Pharmalyzer - ADME and Toxicity Screening Toolkit

Author: [Sorour Hassani]
Email: s.hassani@alum.semnan.ac.ir & sorour.hasani@gmail.com
License: MIT
"""
from rdkit import Chem

TOXIC_SUBSTRUCTURES = {
    "Nitro group": "[N+](=O)[O-]",
    "Azo group": "N=N",
    "Nitrile": "C#N",
    "Aromatic amine": "c1ccc(N)cc1",
    "Hydrazine": "NN",
    "Halogen (Cl, Br, I)": "[Cl,Br,I]",
    "Epoxide": "C1OC1",
    "Acyl chloride": "C(=O)Cl",
    "Isocyanate": "N=C=O",
    "Thiol": "[SH]",
    "Diazo": "N=[N+]",
}

def check_toxic_substructures(smiles):
    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        return None
    found = []
    for name, smarts in TOXIC_SUBSTRUCTURES.items():
        pattern = Chem.MolFromSmarts(smarts)
        if mol.HasSubstructMatch(pattern):
            found.append(name)
    return found

def filter_toxic_compounds(data, smiles_key='smiles'):
    filtered_data = []
    for row in data:
        smiles = row.get(smiles_key, "")
        toxic_features = check_toxic_substructures(smiles)
        if toxic_features is None:
            print(f"Warning: Invalid SMILES '{smiles}'")
            continue
        if not toxic_features:
            filtered_data.append(row)
        else:
            print(f"Removed compound {smiles} due to toxic groups: {', '.join(toxic_features)}")
    return filtered_data




"""
Pharmalyzer - ADME and Toxicity Screening Toolkit

Author: [Sorour Hassani]
Email: s.hassani@alum.semnan.ac.ir & sorour.hasani@gmail.com
License: MIT
"""
def is_missing(val):
    return val in (None, '', 'None', 'null', 'NaN')

def try_parse(val):
    if is_missing(val):
        return None
    try:
        return float(val)
    except ValueError:
        return val

def fill_missing_with_mean(data, columns):
    for col in columns:
        total = 0
        count = 0
        for row in data:
            val = row.get(col)
            if not is_missing(val):
                total += val
                count += 1
        mean = total / count if count else 0
        for row in data:
            if is_missing(row.get(col)):
                row[col] = mean
    return data

def fill_missing_with_value(data, columns, default='Unknown'):
    for col in columns:
        for row in data:
            if is_missing(row.get(col)):
                row[col] = default
    return data

def drop_missing_rows(data, columns):
    """Remove rows where any of the specified columns have missing values."""
    return [row for row in data if all(not is_missing(row.get(col)) for col in columns)]


def fill_missing_forward(data, columns):
    for col in columns:
        last_valid = None
        for row in data:
            if not is_missing(row.get(col)):
                last_valid = row[col]
            elif last_valid is not None:
                row[col] = last_valid
    return data

def fill_missing_backward(data, columns):
    for col in columns:
        next_valid = None
        for row in reversed(data):
            if not is_missing(row.get(col)):
                next_valid = row[col]
            elif next_valid is not None:
                row[col] = next_valid
    return data


from collections import Counter


def fill_missing_with_mode(data, columns):
    """Fill missing values in the specified columns with the most frequent value."""
    for col in columns:
        # Get all non-missing values in the column
        non_missing_values = [row[col] for row in data if not is_missing(row.get(col))]

        # If there are non-missing values, find the mode (most frequent)
        if non_missing_values:
            # Find the most frequent value
            most_frequent_value = Counter(non_missing_values).most_common(1)[0][0]

            # Fill missing values with the most frequent value
            for row in data:
                if is_missing(row.get(col)):
                    row[col] = most_frequent_value
    return data


import math


def euclidean_distance(row1, row2, columns):
    """Calculate Euclidean distance between two rows based on selected columns."""
    distance = 0
    for col in columns:
        if not is_missing(row1.get(col)) and not is_missing(row2.get(col)):
            distance += (row1[col] - row2[col]) ** 2
    return math.sqrt(distance)


def knn_impute(data, columns, k=3):
    """Impute missing values using K-Nearest Neighbors."""
    for col in columns:
        # Find rows with missing values
        rows_with_missing = [row for row in data if is_missing(row.get(col))]

        # Find rows without missing values (for nearest neighbors)
        rows_without_missing = [row for row in data if not is_missing(row.get(col))]

        for row in rows_with_missing:
            # List to store distances and corresponding values of neighbors
            distances = []
            for neighbor in rows_without_missing:
                # Calculate distance between the row with missing value and the neighbor
                distance = euclidean_distance(row, neighbor, columns)
                distances.append((distance, neighbor))

            # Sort the neighbors by distance and get the k-nearest neighbors
            distances.sort(key=lambda x: x[0])
            nearest_neighbors = distances[:k]

            # Take the average (or most common) of the k-nearest neighbors' values for the missing value
            neighbor_values = [neighbor[1][col] for _, neighbor in nearest_neighbors if
                               not is_missing(neighbor[1].get(col))]
            if neighbor_values:
                row[col] = sum(neighbor_values) / len(neighbor_values)  # Use the mean of neighbors
    return data


def clean_mistyped_values(data, column, correction_dict):
    """
    Correct common misspellings or inconsistencies in a categorical column.

    Parameters:
    - data: List of dictionaries (rows).
    - column: The column name where data might have inconsistencies (e.g., 'sex').
    - correction_dict: A dictionary of mappings where keys are incorrect values and values are the correct value.
    """
    for row in data:
        if not is_missing(row.get(column)):
            value = row[column]
            # If the value is in the correction dictionary, update it
            if value in correction_dict:
                row[column] = correction_dict[value]
    return data


from fuzzywuzzy import process


def clean_with_fuzzy_matching(data, column, choices, threshold=80):
    """
    Fix values using fuzzy matching. Values below a certain threshold will not be corrected.

    Parameters:
    - data: List of dictionaries (rows).
    - column: The column name where inconsistencies are to be fixed.
    - choices: List of valid categories (e.g., ['male', 'female']).
    - threshold: Minimum match score for fuzzy matching to consider a correction.
    """
    for row in data:
        if not is_missing(row.get(column)):
            value = row[column]
            # Find the best match in 'choices' using fuzzy matching
            match, score = process.extractOne(value, choices)
            # If the score is above the threshold, correct the value
            if score >= threshold:
                row[column] = match
    return data


def drop_na(data):
    """
    Remove missing values from a list.

    Args:
        data (list): A list of values.

    Returns:
        list: Filtered list without missing values.
    """
    return [x for x in data if not is_missing(x)]


"""
Pharmalyzer - ADME and Toxicity Screening Toolkit

Author: [Sorour Hassani]
Email: s.hassani@alum.semnan.ac.ir & sorour.hasani@gmail.com
License: MIT
"""

from rdkit import Chem, DataStructs
from rdkit.Chem import AllChem

def compute_tanimoto(smiles1, smiles2):
    mol1 = Chem.MolFromSmiles(smiles1)
    mol2 = Chem.MolFromSmiles(smiles2)
    if mol1 is None or mol2 is None:
        return None
    fp1 = AllChem.GetMorganFingerprintAsBitVect(mol1, radius=2, nBits=2048)
    fp2 = AllChem.GetMorganFingerprintAsBitVect(mol2, radius=2, nBits=2048)

    Similarity = DataStructs.TanimotoSimilarity(fp1, fp2)
    return Similarity




"""
Pharmalyzer - ADME and Toxicity Screening Toolkit

Author: [Sorour Hassani]
Email: s.hassani@alum.semnan.ac.ir & sorour.hasani@gmail.com
License: MIT
"""


def physicochemical_properties(smiles):
    mol = Chem.MolFromSmiles(smiles)
    return {
        "MolecularWeight": round(Descriptors.MolWt(mol), 2),
        "LogP": round(Descriptors.MolLogP(mol), 2),
        "TPSA": round(rdMolDescriptors.CalcTPSA(mol), 2),
        "HBD": Descriptors.NumHDonors(mol),
        "HBA": Descriptors.NumHAcceptors(mol),
        "RotatableBonds": Descriptors.NumRotatableBonds(mol),
        "Rings": mol.GetRingInfo().NumRings(),
        "HeavyAtoms": mol.GetNumHeavyAtoms()
    }


from rdkit import Chem
from rdkit.Chem import Descriptors, rdMolDescriptors

def pharmacokinetics_properties(smiles):
    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        return {"GI_Absorption": "Unknown", "BBB_Permeability": "Unknown"}

    logp = Descriptors.MolLogP(mol)  # Approximate WLOGP
    tpsa = rdMolDescriptors.CalcTPSA(mol)

    # GI Absorption
    if tpsa <= 131.6 and (-0.7 <= logp <= 5.0):
        gi = "High"
    else:
        gi = "Low"

    # BBB Permeability
    if tpsa <= 90 and (-0.7 <= logp <= 5.0):
        bbb = "Yes"
    else:
        bbb = "No"

    return {"GI_Absorption": gi, "BBB_Permeability": bbb}


def filter_by_tpsa(data, column='tpsa', max_val=140):

    filtered = []
    for row in data:
        try:
            tpsa = float(row.get(column, 0))
            if tpsa <= max_val:
                filtered.append(row)
        except:
            continue
    return filtered

def filter_by_rotatable_bonds(data, column='rotatable_bonds', max_val=10):
    filtered = []
    for row in data:
        try:
            rb = int(row.get(column, 0))
            if rb <= max_val:
                filtered.append(row)
        except:
            continue
    return filtered

def filter_by_logS(data, column='logS', min_val=-6):
    filtered = []
    for row in data:
        try:
            logs = float(row.get(column, 0))
            if logs >= min_val:
                filtered.append(row)
        except:
            continue
    return filtered


def classify_logS_adme(logs):
    try:
        logs = float(logs)
        if logs > 0:
            return "Very Soluble"
        elif 0 >= logs > -2:
            return "Soluble"
        elif -2 >= logs > -4:
            return "Moderately Soluble"
        elif -4 >= logs > -6:
            return "Poorly Soluble"
        elif -6 >= logs > -10:
            return "Very Poorly Soluble"
        elif logs <= -10:
            return "Insoluble"
    except (ValueError, TypeError):
        return "Invalid logS"



def clean_adme_compounds(data, tpsa_max=140, rb_max=10, logs_min=-6):
    step1 = filter_by_tpsa(data, max_val=tpsa_max)
    step2 = filter_by_rotatable_bonds(step1, max_val=rb_max)
    step3 = filter_by_logS(step2, min_val=logs_min)
    return step3

from rdkit import Chem
from rdkit.Chem import Descriptors

def compute_logp(smiles: str) -> float:
    try:
        mol = Chem.MolFromSmiles(smiles)
        if mol is None:
            return None
        return Descriptors.MolLogP(mol)
    except Exception as e:
        print(f"Error computing logP: {e}")
        return None


from rdkit import Chem
from rdkit.Chem import FilterCatalog
from rdkit.Chem.FilterCatalog import FilterCatalogParams

def contains_pains(smiles):
    """
    Check if a molecule contains PAINS substructures.
    Returns True if PAINS found, otherwise False.
    """
    mol = Chem.MolFromSmiles(smiles)
    if not mol:
        return False

    params = FilterCatalogParams()
    params.AddCatalog(FilterCatalogParams.FilterCatalogs.PAINS_A)
    params.AddCatalog(FilterCatalogParams.FilterCatalogs.PAINS_B)
    params.AddCatalog(FilterCatalogParams.FilterCatalogs.PAINS_C)

    catalog = FilterCatalog.FilterCatalog(params)

    return catalog.HasMatch(mol)

def filter_pains(data, smiles_column='smiles'):
    """
    Remove molecules with PAINS substructures from the dataset.
    :param data: list of dicts (e.g., from CSV or JSON)
    :param smiles_column: column name containing SMILES strings
    :return: filtered list
    """
    filtered = []
    for row in data:
        smiles = row.get(smiles_column, '')
        if smiles and not contains_pains(smiles):
            filtered.append(row)
    return filtered

from rdkit import Chem
from rdkit.Chem import FilterCatalog
from rdkit.Chem.FilterCatalog import FilterCatalogParams

def contains_brenk(smiles):
    """
    Check if a molecule contains Brenk substructures.
    Returns True if matched, False otherwise.
    """
    mol = Chem.MolFromSmiles(smiles)
    if not mol:
        return False

    params = FilterCatalogParams()
    params.AddCatalog(FilterCatalogParams.FilterCatalogs.Brenk)
    catalog = FilterCatalog.FilterCatalog(params)

    return catalog.HasMatch(mol)

def filter_brenk(data, smiles_column='smiles'):
    """
    Remove molecules that contain Brenk alerts from the dataset.
    :param data: list of dicts (CSV or JSON rows)
    :param smiles_column: name of SMILES column
    :return: filtered data list
    """
    filtered = []
    for row in data:
        smiles = row.get(smiles_column, '')
        if smiles and not contains_brenk(smiles):
            filtered.append(row)
    return filtered


from rdkit import Chem
from rdkit.Chem import Descriptors, Crippen, Lipinski, rdMolDescriptors

def is_lead_like(smiles):
    mol = Chem.MolFromSmiles(smiles)
    if not mol:
        return False

    mw = Descriptors.MolWt(mol)
    logp = Crippen.MolLogP(mol)
    h_donors = Lipinski.NumHDonors(mol)
    h_acceptors = Lipinski.NumHAcceptors(mol)
    rot_bonds = Lipinski.NumRotatableBonds(mol)
    tpsa = rdMolDescriptors.CalcTPSA(mol)

    return (
        250 <= mw <= 350 and
        logp <= 3.5 and
        h_donors <= 3 and
        h_acceptors <= 6 and
        rot_bonds <= 7 and
        tpsa <= 90
    )

def filter_lead_like(data, smiles_column='smiles'):
    filtered = []
    for row in data:
        smiles = row.get(smiles_column, '')
        if smiles and is_lead_like(smiles):
            filtered.append(row)
    return filtered


from rdkit import Chem
from rdkit.Chem import Descriptors

def compute_log_kp(smiles):
    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        return None, "Invalid SMILES"

    logP = Descriptors.MolLogP(mol)
    mw = Descriptors.MolWt(mol)

    log_kp = -2.72 + 0.71 * logP - 0.0061 * mw
    return round(log_kp, 4), classify_log_kp(log_kp)

def classify_log_kp(log_kp):
    if log_kp > -2:
        return "High permeability"
    elif -4 < log_kp <= -2:
        return "Moderate permeability"
    else:
        return "Low permeability"

from rdkit import Chem
from rdkit.Chem import Descriptors, rdMolDescriptors

def predict_excretion(smiles):
    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        return "Invalid SMILES"

    mw = Descriptors.MolWt(mol)
    logp = Descriptors.MolLogP(mol)
    tpsa = rdMolDescriptors.CalcTPSA(mol)

    if mw < 400 and logp < 3 and tpsa > 75:
        return "Likely Renal (kidney) Excretion"
    elif logp > 4 and tpsa < 60:
        return "Likely Biliary (liver/bile) Excretion"
    else:
        return "Mixed or Uncertain Excretion Pathway"

from rdkit import Chem
from rdkit.Chem import Descriptors, rdMolDescriptors

def cyp_inhibition_prediction(smiles):
    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        return "Invalid SMILES"

    logp = Descriptors.MolLogP(mol)
    tpsa = rdMolDescriptors.CalcTPSA(mol)
    hbd = Descriptors.NumHDonors(mol)
    hba = Descriptors.NumHAcceptors(mol)
    mw = Descriptors.MolWt(mol)
    aromatic_rings = len([r for r in mol.GetRingInfo().AtomRings()
                          if all(mol.GetAtomWithIdx(i).GetIsAromatic() for i in r)])

    prediction = {}

    # Rule-based estimations (based on literature & SwissADME tendencies)
    prediction['CYP1A2'] = "Likely Inhibitor" if logp > 2.5 and tpsa < 75 and aromatic_rings >= 2 else "Unlikely"
    prediction['CYP2C9'] = "Likely Inhibitor" if logp > 2.0 and mw > 300 else "Unlikely"
    prediction['CYP2C19'] = "Likely Inhibitor" if logp > 3.0 and hbd <= 2 else "Unlikely"
    prediction['CYP2D6'] = "Likely Inhibitor" if logp > 3.0 and hba >= 3 else "Unlikely"
    prediction['CYP3A4'] = "Likely Inhibitor" if logp > 3.5 and hbd <= 2 and mw > 350 else "Unlikely"

    return prediction
def cyp_substrate_prediction(smiles):
    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        return "Invalid SMILES"

    logp = Descriptors.MolLogP(mol)
    mw = Descriptors.MolWt(mol)
    rot_bonds = Descriptors.NumRotatableBonds(mol)
    hbd = Descriptors.NumHDonors(mol)
    tpsa = rdMolDescriptors.CalcTPSA(mol)

    result = {}

    # Based on general trends (substrates often moderate size, moderate LogP, flexible)
    result['CYP3A4'] = "Likely Substrate" if mw > 350 and logp > 2.5 and rot_bonds > 5 else "Unlikely"
    result['CYP2D6'] = "Likely Substrate" if logp > 2.0 and hbd >= 1 else "Unlikely"
    result['CYP2C9'] = "Likely Substrate" if mw > 300 and tpsa < 90 else "Unlikely"
    result['CYP1A2'] = "Likely Substrate" if logp > 2.0 and tpsa < 70 else "Unlikely"

    return result



"""
Pharmalyzer - ADME and Toxicity Screening Toolkit

Author: [Sorour Hassani]
Email: s.hassani@alum.semnan.ac.ir & sorour.hasani@gmail.com
License: MIT
"""
import numpy as np
import pandas as pd

#Outlier Detection using Z-score (for normal distribution)

import numpy as np


def detect_outliers_zscore(data, columns, threshold=3):
    """
    Detects outliers using the Z-score method for the specified columns.
    :param data: DataFrame
    :param columns: List of columns to check for outliers
    :param threshold: Z-score value above which data will be considered as an outlier
    :return: DataFrame with outliers replaced by NaN
    """
    for column in columns:
        mean = np.mean(data[column])
        std_dev = np.std(data[column])
        z_scores = [(x - mean) / std_dev for x in data[column]]
        data[column] = [x if abs(z) <= threshold else np.nan for x, z in zip(data[column], z_scores)]

    return data

#Outlier Detection using IQR (for skewed distributions)

def detect_outliers_iqr(data, columns, lower_quantile=0.25, upper_quantile=0.75, factor=1.5):
    """
    Detects outliers using the IQR (Interquartile Range) method.
    :param data: DataFrame
    :param columns: List of columns to check for outliers
    :param lower_quantile: Lower quartile (default is 0.25)
    :param upper_quantile: Upper quartile (default is 0.75)
    :param factor: Multiplier for determining the outlier threshold (default is 1.5)
    :return: DataFrame with outliers replaced by NaN
    """
    for column in columns:
        Q1 = data[column].quantile(lower_quantile)
        Q3 = data[column].quantile(upper_quantile)
        IQR = Q3 - Q1
        lower_bound = Q1 - (IQR * factor)
        upper_bound = Q3 + (IQR * factor)
        data[column] = [x if lower_bound <= x <= upper_bound else np.nan for x in data[column]]

    return data

#Cap or Clip Outliers

def cap_outliers(data, columns, lower_percentile=0.05, upper_percentile=0.95):
    """
    Caps outliers in the specified columns by replacing them with the values from the specified percentiles.
    :param data: DataFrame
    :param columns: List of columns to apply capping
    :param lower_percentile: Lower percentile (default is 5th percentile)
    :param upper_percentile: Upper percentile (default is 95th percentile)
    :return: DataFrame with capped outliers
    """
    for column in columns:
        lower = data[column].quantile(lower_percentile)
        upper = data[column].quantile(upper_percentile)
        data[column] = data[column].clip(lower=lower, upper=upper)

    return data

#Replace Outliers with the Median
def replace_outliers_with_median(data, columns, threshold=3):
    """
    Replaces outliers with the median of each column.
    :param data: DataFrame
    :param columns: List of columns to check for outliers
    :param threshold: Z-score value above which data will be considered as an outlier
    :return: DataFrame with outliers replaced by the column median
    """
    for column in columns:
        mean = np.mean(data[column])
        std_dev = np.std(data[column])
        z_scores = [(x - mean) / std_dev for x in data[column]]
        median = np.median(data[column])
        data[column] = [median if abs(z) > threshold else x for x, z in zip(data[column], z_scores)]

    return data




"""
Pharmalyzer - ADME and Toxicity Screening Toolkit

Author: [Sorour Hassani]
Email: s.hassani@alum.semnan.ac.ir & sorour.hasani@gmail.com
License: MIT
"""
import pandas as pd

def remove_non_numerical_columns(data):
    """
    Removes non-numeric columns from the dataset.
    """
    return data.select_dtypes(include=['number'])

def drop_missing_threshold(data, col_thresh=0.5, row_thresh=0.5):
    """
    Drops columns and rows with missing data above a threshold.
    - col_thresh: minimum fraction of non-null values a column must have.
    - row_thresh: minimum fraction of non-null values a row must have.
    """
    data = data.dropna(axis=1, thresh=int(col_thresh * len(data)))
    data = data.dropna(axis=0, thresh=int(row_thresh * len(data.columns)))
    return data

def drop_discrete_columns(data, max_unique=5):
    """
    Drops categorical columns with low cardinality (discrete columns).
    """
    to_drop = [col for col in data.columns
               if data[col].dtype == 'object' and data[col].nunique() <= max_unique]
    return data.drop(columns=to_drop)

def drop_constant_columns(data):
    """
    Drops columns with the same value in all rows (zero variance).
    """
    return data.loc[:, data.nunique() > 1]

def drop_index_columns(data):
    """
    Drops columns that are likely to be index or case number identifiers.
    """
    likely_ids = [col for col in data.columns
                  if 'id' in col.lower() or 'case' in col.lower()]
    return data.drop(columns=likely_ids)

def clean_dataset(data, col_thresh=0.5, row_thresh=0.5, drop_discrete=True):
    """
    Runs a full cleaning pipeline:
    - Removes columns/rows with high missingness
    - Removes constant columns
    - Removes ID/case-like columns
    - Removes non-numeric columns
    - Optionally drops low-cardinality discrete columns
    """
    data = drop_missing_threshold(data, col_thresh, row_thresh)
    data = drop_constant_columns(data)
    data = drop_index_columns(data)
    data = remove_non_numerical_columns(data)
    if drop_discrete:
        data = drop_discrete_columns(data)
    return data






"""
Pharmalyzer - ADME and Toxicity Screening Toolkit

Author: [Sorour Hassani]
Email: s.hassani@alum.semnan.ac.ir & sorour.hasani@gmail.com
License: MIT
"""
def check_lipinski_rule(data):
    passed = []
    failed = []

    for row in data:
        try:
            mw = float(row.get("molecular_weight", 0))
            logp = float(row.get("logp", 0))
            h_donors = int(row.get("h_donors", 0))
            h_acceptors = int(row.get("h_acceptors", 0))

            if (
                mw <= 500 and
                logp <= 5 and
                h_donors <= 5 and
                h_acceptors <= 10
            ):
                passed.append(row)
            else:
                failed.append(row)

        except (ValueError, TypeError):
            continue

    return passed, failed


def filter_by_rule(data, rule_func):
    """
    Apply a rule function and return only the passing compounds.

    Parameters:
        data (list of dict)
        rule_func (function): A function like check_lipinski_rule

    Returns:
        list of dict: Compounds passing the rule
    """
    passed, _ = rule_func(data)
    return passed



def check_veber_rule(data):
    """
    Check Veber's rule for drug-likeness:
    - Rotatable bonds ≤ 10
    - Polar Surface Area (PSA) ≤ 140

    Parameters:
        data (list of dict)

    Returns:
        tuple: (passed, failed)
    """
    passed = []
    failed = []

    for row in data:
        try:
            rb = int(row.get("rotatable_bonds", 0))
            psa = float(row.get("psa", 0))

            if rb <= 10 and psa <= 140:
                passed.append(row)
            else:
                failed.append(row)
        except (ValueError, TypeError):
            continue

    return passed, failed



def check_egan_rule(data):
    """
    Check Egan's rule:
    - logP ≤ 5.88
    - PSA ≤ 131.6

    Parameters:
        data (list of dict)

    Returns:
        tuple: (passed, failed)
    """
    passed = []
    failed = []

    for row in data:
        try:
            logp = float(row.get("logp", 0))
            psa = float(row.get("psa", 0))

            if logp <= 5.88 and psa <= 131.6:
                passed.append(row)
            else:
                failed.append(row)
        except (ValueError, TypeError):
            continue

    return passed, failed


def check_muegge_rule(row):
    try:
        mw = float(row.get("molecular_weight", 0))
        logp = float(row.get("logp", 0))
        tpsa = float(row.get("tpsa", 0))
        rings = int(row.get("ring_count", 0))
        carbon = int(row.get("carbon_count", 0))

        return (200 <= mw <= 600 and
                -2 <= logp <= 5 and
                tpsa <= 150 and
                rings <= 7 and
                carbon >= 4)
    except (ValueError, TypeError):
        return False


def check_ghose_filter(row):
    try:
        mw = float(row.get("molecular_weight", 0))
        logp = float(row.get("logp", 0))
        atoms = int(row.get("atom_count", 0))
        refractivity = float(row.get("molar_refractivity", 0))

        return (160 <= mw <= 480 and
                0.13 <= logp <= 5.6 and
                40 <= atoms <= 70 and
                20 <= refractivity <= 130)
    except (ValueError, TypeError):
        return False



def evaluate_drug_rules(data):
    results = []

    for row in data:
        compound_id = row.get("id", "unknown")

        result = {
            "id": compound_id,
            "Lipinski": check_lipinski_rule(row),
            "Ghose": check_ghose_filter(row),
            "Veber": check_veber_rule(row),
            "Egan": check_egan_rule(row),
            "Muegge": check_muegge_rule(row)
        }

        results.append(result)

    return results






"""
Pharmalyzer - ADME and Toxicity Screening Toolkit

Author: [Sorour Hassani]
Email: s.hassani@alum.semnan.ac.ir & sorour.hasani@gmail.com
License: MIT
"""
# Pharmalyzer/chembl_client.py

from chembl_webresource_client.new_client import new_client

class ChEMBLClient:
    """
    Class for accessing ChEMBL data via the official API.
    """

    def __init__(self):
        # Initialize molecule and activity clients
        self.molecule_client = new_client.molecule
        self.activity_client = new_client.activity

    def search_molecules(self, query, page_size=20):
        """
        Search for molecules based on name, InChIKey, or other fields.
        Returns a list of dictionaries with molecule information.

        Args:
            query (str): The search term (e.g., part of a molecule name or ID).
            page_size (int): Number of results to return (default 20).

        Returns:
            list: A list of dicts containing 'molecule_chembl_id', 'canonical_smiles', and 'pref_name'.
        """
        results = (
            self.molecule_client
            .filter(molecule_chembl_id__icontains=query)
            .only(['molecule_chembl_id', 'canonical_smiles', 'pref_name'])
            .order_by('molecule_chembl_id')
            .limit(page_size)
        )
        return list(results)

    def fetch_activity_by_target(self, target_chembl_id, activity_types=None, limit=100):
        """
        Retrieve activity data (e.g., IC50, Ki) for a specific target (e.g., CHEMBL25).
        Returns a list of dictionaries containing SMILES, activity type, value, and unit.

        Args:
            target_chembl_id (str): ChEMBL ID of the target (e.g., 'CHEMBL25').
            activity_types (str or list of str, optional): Activity type(s) to filter (e.g., "IC50", "Ki", or ["IC50", "Ki"]).
                                                             If None, retrieves all activity types.
            limit (int): Maximum number of records to retrieve (default 100).

        Returns:
            list: A list of dicts with keys 'smiles', 'activity_type', 'activity_value', and 'activity_unit'.
        """
        # If activity_types is a single string, convert it to a list
        if isinstance(activity_types, str):
            activity_types = [activity_types]

        # Start filtering by target
        query = self.activity_client.filter(target_chembl_id=target_chembl_id)

        # If specific activity types are requested, apply the 'standard_type__in' filter
        if activity_types:
            query = query.filter(standard_type__in=activity_types)

        # Retrieve only the desired fields
        records = (
            query
            .only(['canonical_smiles', 'standard_type', 'standard_value', 'standard_units'])
            .order_by('standard_value')
            .limit(limit)
        )

        dataset = []
        for entry in records:
            smiles = entry.get('canonical_smiles')
            act_type = entry.get('standard_type')
            value = entry.get('standard_value')
            unit = entry.get('standard_units')

            # Keep only entries with valid SMILES and numeric value, and accepted units
            if smiles and value and unit and unit.lower() in ['nm', 'µm', 'mm']:
                try:
                    value = float(value)
                except (ValueError, TypeError):
                    continue
                dataset.append({
                    'smiles': smiles,
                    'activity_type': act_type,
                    'activity_value': float(value),
                    'activity_unit': unit
                })

        return dataset

    def fetch_molecule_by_chembl_id(self, chembl_id):
        """
        Retrieve full information for a molecule based on its ChEMBL ID (e.g., 'CHEMBL25').

        Args:
            chembl_id (str): The ChEMBL ID of the molecule.

        Returns:
            dict: A dictionary containing detailed molecule data.
        """
        return self.molecule_client.get(chembl_id)



__all__ = ['fill_missing_with_mean', 'fill_missing_with_value', 'label_encode', 'min_max_scale','drop_missing_rows','fill_missing_forward','fill_missing_backward',
           'fill_missing_with_mode','knn_impute','clean_mistyped_values','clean_with_fuzzy_matching',
           'detect_outliers_zscore','detect_outliers_iqr','cap_outliers','replace_outliers_with_median',
           'clean_dataset','remove_non_numerical_columns','drop_missing_threshold','drop_constant_columns',
           'drop_discrete_columns','drop_index_columns','integrate_matrices_with_header','check_lipinski_rule',
           'filter_by_rule','check_veber_rule','check_egan_rule','check_muegge_rule','check_ghose_filter',
           'evaluate_drug_rules','clean_adme_compounds','filter_by_tpsa','classify_logS_adme','filter_by_rotatable_bonds',
           'filter_toxic_compounds','check_toxic_substructures','compute_logp','filter_pains','filter_brenk',
           'filter_lead_like','physicochemical_properties','pharmacokinetics_properties','compute_log_kp',
           'predict_excretion','cyp_inhibition_prediction','cyp_substrate_prediction','ChEMBLClient','min_max_scale_multiple',
           'filter_by_logS','drop_na']

from .cleaner import fill_missing_with_mean, fill_missing_with_value
from .encoder import label_encode
from .scaler import min_max_scale
from .scaler import min_max_scale_multiple
from .cleaner import fill_missing_forward, fill_missing_backward
from .cleaner import drop_missing_rows
from .cleaner import fill_missing_with_mode
from .cleaner import knn_impute
from .cleaner import clean_mistyped_values, clean_with_fuzzy_matching
from .outliers import detect_outliers_zscore, detect_outliers_iqr, cap_outliers, replace_outliers_with_median
from .filtering import (
    clean_dataset,
    remove_non_numerical_columns,
    drop_missing_threshold,
    drop_discrete_columns,
    drop_constant_columns,
    drop_index_columns
)
from.integrate import integrate_matrices_with_header
from .drug_rules import check_lipinski_rule
from .drug_rules import check_lipinski_rule, filter_by_rule
from .drug_rules import check_veber_rule, check_egan_rule
from .drug_rules import check_muegge_rule
from .drug_rules import check_ghose_filter
from .drug_rules import evaluate_drug_rules
from .ADME import filter_by_tpsa
from .ADME import filter_by_rotatable_bonds
from .ADME import classify_logS_adme
from .ADME import clean_adme_compounds
from .toxicity import check_toxic_substructures
from .toxicity import filter_toxic_compounds
from .ADME import compute_logp
from .ADME import filter_pains
from .ADME import filter_brenk
from .ADME import filter_lead_like
from .ADME import physicochemical_properties
from .ADME import pharmacokinetics_properties
from .ADME import compute_log_kp
from .ADME import predict_excretion
from .ADME import cyp_inhibition_prediction
from .ADME import cyp_substrate_prediction
from .ADME import  filter_by_logS
from .chembl_client import ChEMBLClient
from .qsar import QSARModel
from .similarity import compute_tanimoto
from .cleaner import drop_na



from setuptools import setup, find_packages

setup(
    name='Pharmalyzer',
    version='0.1',
    author='[Sorour Hassani]',
    author_email='s.hassani@alum.semnan.ac.ir & sorour.hasani@gmail.com',
    description='Pharmaceutical ADME/Tox filtering and analysis toolkit',
    packages=find_packages(),
    license='MIT',
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
    install_requires=[
        'rdkit',
        'fuzzywuzzy',
        'python-Levenshtein'
    ],
)


"""
Pharmalyzer - ADME and Toxicity Screening Toolkit

Author: [Sorour Hassani]
Email: s.hassani@alum.semnan.ac.ir & sorour.hasani@gmail.com
License: MIT
"""
def label_encode(data, column):
    unique_vals = list({row[column] for row in data if row.get(column) is not None})
    mapping = {val: idx for idx, val in enumerate(sorted(unique_vals))}
    for row in data:
        if row.get(column) in mapping:
            row[column] = mapping[row[column]]
        else:
            row[column] = -1
    return data, mapping



"""
Pharmalyzer - ADME and Toxicity Screening Toolkit

Author: [Sorour Hassani]
Email: s.hassani@alum.semnan.ac.ir & sorour.hasani@gmail.com
License: MIT
"""

def integrate_matrices_with_header(matrix1, headers1, matrix2, headers2, base_on='headers1'):
    """
    ترکیب دو ماتریس بر اساس ترتیب ستون‌های مشخص شده.

    پارامترها:
        matrix1, headers1: ماتریس و هدر اول
        matrix2, headers2: ماتریس و هدر دوم
        base_on (str): 'headers1' یا 'headers2' برای مشخص کردن مبنای ترتیب ستون

    خروجی:
        combined_matrix, final_headers
    """
    if base_on not in ['headers1', 'headers2']:
        raise ValueError("base_on must be 'headers1' or 'headers2'")

    final_headers = headers1 if base_on == 'headers1' else headers2

    def reorder_matrix(matrix, original_headers):
        index_map = {}
        for col in final_headers:
            if col in original_headers:
                index_map[col] = original_headers.index(col)
            else:
                raise ValueError(f"Column '{col}' not found in headers: {original_headers}")

        reordered = []
        for row in matrix:
            reordered.append([row[index_map[col]] if col in index_map else None for col in final_headers])
        return reordered

    matrix1_reordered = reorder_matrix(matrix1, headers1)
    matrix2_reordered = reorder_matrix(matrix2, headers2)

    combined_matrix = matrix1_reordered + matrix2_reordered

    return combined_matrix, final_headers


